@@ -1,6 +1,6 @@
 /* crlcache.c - LDAP access
- *      Copyright (C) 2002 Klarlvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004 g10 Code GmbH
+ *      Copyright (C) 2002 Klarälvdalens Datakonsult AB
+ *      Copyright (C) 2003, 2004, 2005 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -97,20 +97,22 @@
 #include <sys/utsname.h>
 
 #include "dirmngr.h"
+#include "validate.h"
+#include "certcache.h"
 #include "crlcache.h"
 #include "crlfetch.h"
 #include "misc.h"
 #include "cdb.h"
 
 /* Change this whenever the format changes */
-#define DBDIR_D "dirmngr-cache.d"
+#define DBDIR_D (opt.system_daemon? "crls.d" : "dirmngr-cache.d")
 #define DBDIRFILE "DIR.txt"
 #define DBDIRVERSION 1
 
 /* The number of DB files we may have open at one time.  We need to
    limit this because there is no guarantee that the number of issuers
    has a upper limit.  We are currently using mmap, so it is a good
-   idea anyway to limit the number of opend cache files. */
+   idea anyway to limit the number of opened cache files. */
 #define MAX_OPEN_DB_FILES 5
 
 
@@ -154,12 +156,34 @@ struct crl_cache_s
   unsigned int cdb_open_count; /* Currently open cache files. */
 };
 
+typedef struct crl_cache_s *crl_cache_t;
 
+
+/* Prototypes.  */
 static crl_cache_entry_t find_entry (crl_cache_entry_t first,
                                      const char *issuer_hash);
 
 
 
+/* The currently loaded cache object.  This isi usually initialized
+   right at startup.  */
+static crl_cache_t current_cache;
+
+
+
+
+
+/* Return the current cache object or bail out if it is has not yet
+   been initialized.  */
+static crl_cache_t
+get_current_cache (void)
+{
+  if (!current_cache)
+    log_fatal ("CRL cache has not yet been initialized\n");
+  return current_cache;
+}
+
+
 /* 
    Create ae directory if it does not yet exists.  Returns on
    success, or -1 on error.
@@ -170,7 +194,7 @@ create_directory_if_needed (const char *name)
   DIR *dir;
   char *fname;
 
-  fname = make_filename (opt.homedir, name, NULL);
+  fname = make_filename (opt.homedir_cache, name, NULL);
   dir = opendir (fname);
   if (!dir)
     {
@@ -197,7 +221,7 @@ create_directory_if_needed (const char *name)
 static int
 cleanup_cache_dir (int force)
 {
-  char *dname = make_filename (opt.homedir, DBDIR_D, NULL);
+  char *dname = make_filename (opt.homedir_cache, DBDIR_D, NULL);
   DIR *dir;
   struct dirent *de;
   int problem = 0;
@@ -449,7 +473,7 @@ check_dir_version (FILE **fpadr, const char *fname,
 
 
 /* Open the dir file and read in all available information.  Store
-   that in a newly allocated cahce object and return that if
+   that in a newly allocated cache object and return that if
    everything worked out fine.  Create the cache directory and the dir
    if it does not yet exist.  Remove all files in that directory if
    the version does not match. */
@@ -465,7 +489,7 @@ open_dir (crl_cache_t *r_cache)
   gpg_error_t err = 0;
   int anyerr = 0;
 
-  fname = make_filename (opt.homedir, DBDIR_D, DBDIRFILE, NULL);
+  fname = make_filename (opt.homedir_cache, DBDIR_D, DBDIRFILE, NULL);
 
   lineno = 0;
   fp = open_dir_file (fname);
@@ -663,7 +687,7 @@ update_dir (crl_cache_t cache)
   unsigned int lineno;
   gpg_error_t err = 0;
 
-  fname = make_filename (opt.homedir, DBDIR_D, DBDIRFILE, NULL);
+  fname = make_filename (opt.homedir_cache, DBDIR_D, DBDIRFILE, NULL);
 
   /* Fixme: Take an update file lock here. */
 
@@ -701,7 +725,7 @@ update_dir (crl_cache_t cache)
     for (p=tmpbuf; *p; p++)
       if (*p == '/')
         *p = '.';
-    tmpfname = make_filename (opt.homedir, DBDIR_D, tmpbuf, NULL);
+    tmpfname = make_filename (opt.homedir_cache, DBDIR_D, tmpbuf, NULL);
     xfree (tmpbuf);
   }
   fpout = fopen (tmpfname, "w");
@@ -835,7 +859,7 @@ make_db_file_name (const char *issuer_hash)
   memcpy (bname, "crl-", 4);
   memcpy (bname + 4, issuer_hash, 40);
   strcpy (bname + 44, ".db");
-  return make_filename (opt.homedir, DBDIR_D, bname, NULL);
+  return make_filename (opt.homedir_cache, DBDIR_D, bname, NULL);
 }
 
 
@@ -1057,31 +1081,38 @@ find_entry (crl_cache_entry_t first, const char *issuer_hash)
 
 
 
-/* Create a new CRL cache. Returns the handle to this cache. Will
+/* Create a new CRL cache. This fucntion is usually called only once.
    never fail. */
-crl_cache_t
+void
 crl_cache_init(void)
 {
   crl_cache_t cache;
   gpg_error_t err;
 
+  if (current_cache)
+    {
+      log_error ("crl cache has already been initialized - not doing twice\n");
+      return;
+    }
+
   err = open_dir (&cache);
-  if (err)
+  if (err) 
     log_fatal (_("failed to create a new cache object: %s\n"),
                gpg_strerror (err));
-  return cache;
+  current_cache = cache;
 }
 
 
 /* Remove the cache information and all its resources.  Note that we
    still keep the cache on disk. */
 void 
-crl_cache_deinit (crl_cache_t cache)
+crl_cache_deinit (void)
 {
-  if (!cache)
-    return;
-
-  release_cache (cache);
+  if (current_cache)
+    {
+      release_cache (current_cache);
+      current_cache = NULL;
+    }
 }
 
 
@@ -1089,32 +1120,31 @@ crl_cache_deinit (crl_cache_t cache)
 int 
 crl_cache_flush (void)
 {
-  int problem = 0;
+  int rc;
 
-  if (cleanup_cache_dir (0))
-    problem = -1;
-  return problem;
+  rc = cleanup_cache_dir (0)? -1 : 0;
+
+  return rc;
 }
 
 
 /* Check whether the certificate identified by ISSUER_HASH and
-   SERIALNO is valid; i.e. not listed in our cache.  With
-   FORCE_REFRESH set to true, a new CRL will be retrieved eben if the
-   cached has not yet expired.  We use a 30 minutes threshold here so
+   SN/SNLEN is valid; i.e. not listed in our cache.  With
+   FORCE_REFRESH set to true, a new CRL will be retrieved even if the
+   cache has not yet expired.  We use a 30 minutes threshold here so
    that invoking this function several times won't load the CRL over
    and over.  */
-crl_cache_result_t 
-crl_cache_isvalid (crl_cache_t cache,
-                   const char *issuer_hash, const char *serialno,
-                   int force_refresh)
+static crl_cache_result_t 
+cache_isvalid (ctrl_t ctrl, const char *issuer_hash,
+               const unsigned char *sn, size_t snlen,
+               int force_refresh)
 {
+  crl_cache_t cache = get_current_cache ();
   crl_cache_result_t retval;
   struct cdb *cdb;
   int rc;
   crl_cache_entry_t entry;
   dirmngr_isotime_t current_time;
-  unsigned char snbuf_buffer[50];
-  unsigned char *snbuf;
   size_t n;
 
   entry = find_entry (cache->entries, issuer_hash);
@@ -1175,38 +1205,39 @@ crl_cache_isvalid (crl_cache_t cache,
       return CRL_CACHE_DONTKNOW;
     }
 
-  n = strlen (serialno)/2+1;
-  if (n < sizeof snbuf_buffer - 1)
-    snbuf = snbuf_buffer;
-  else
-    snbuf = xmalloc (n);
-
-  n = unhexify (snbuf, serialno);
-  rc = cdb_find (cdb, snbuf, n);
+  rc = cdb_find (cdb, sn, snlen);
   if (rc == 1) 
     {
       n = cdb_datalen (cdb);
       if (n != 16)
-        log_error (_("WARNING: invalid cache record length for S/N %s\n"),
-                   serialno );
+        {
+          log_error (_("WARNING: invalid cache record length for S/N "));
+          log_printhex ("", sn, snlen);
+        }
       else if (opt.verbose)
         {
           unsigned char record[16];
+          char *tmp = hexify_data (sn, snlen);
 
           if (cdb_read (cdb, record, n, cdb_datapos (cdb)))
             log_error (_("problem reading cache record for S/N %s: %s\n"),
-                       serialno, strerror (errno));
+                       tmp, strerror (errno));
           else
             log_info (_("S/N %s is not valid; reason=%02X  date=%.15s\n"),
-                      serialno, *record, record+1);
+                      tmp, *record, record+1);
+          xfree (tmp);
         }
       retval = CRL_CACHE_INVALID;
     }
   else if (!rc)
     {
       if (opt.verbose)
-        log_info (_("S/N %s is valid, it is not listed in the CRL\n"),
-                  serialno );
+        {
+          char *serialno = hexify_data (sn, snlen);
+          log_info (_("S/N %s is valid, it is not listed in the CRL\n"),
+                    serialno );
+          xfree (serialno);
+        }
       retval = CRL_CACHE_VALID;
     }
   else 
@@ -1216,66 +1247,207 @@ crl_cache_isvalid (crl_cache_t cache,
       retval = CRL_CACHE_DONTKNOW;
     }
 
-  if (snbuf != snbuf_buffer)
-    xfree (snbuf);
-
   unlock_db_file (cache, entry);
 
   return retval;
 }
 
 
+/* Check whether the certificate identified by ISSUER_HASH and
+   SERIALNO is valid; i.e. not listed in our cache.  With
+   FORCE_REFRESH set to true, a new CRL will be retrieved even if the
+   cache has not yet expired.  We use a 30 minutes threshold here so
+   that invoking this function several times won't load the CRL over
+   and over.  */
+crl_cache_result_t 
+crl_cache_isvalid (ctrl_t ctrl, const char *issuer_hash, const char *serialno,
+                   int force_refresh)
+{
+  crl_cache_result_t result;
+  unsigned char snbuf_buffer[50];
+  unsigned char *snbuf;
+  size_t n;
 
-/* Return the certificate for ISSUER or NULL if it can't be found. */
-static ksba_cert_t 
-get_issuer_cert (ctrl_t ctrl, const char *issuer)
+  n = strlen (serialno)/2+1;
+  if (n < sizeof snbuf_buffer - 1)
+    snbuf = snbuf_buffer;
+  else
+    snbuf = xmalloc (n);
+
+  n = unhexify (snbuf, serialno);
+
+  result = cache_isvalid (ctrl, issuer_hash, snbuf, n, force_refresh);
+
+  if (snbuf != snbuf_buffer)
+    xfree (snbuf);
+
+  return result;
+}  
+
+
+/* Check whether the certificate CERT is valid; i.e. not listed in our
+   cache.  With FORCE_REFRESH set to true, a new CRL will be retrieved
+   even if the cache has not yet expired.  We use a 30 minutes
+   threshold here so that invoking this function several times won't
+   load the CRL over and over.  */
+gpg_error_t
+crl_cache_cert_isvalid (ctrl_t ctrl, ksba_cert_t cert,
+                        int force_refresh)
 {
   gpg_error_t err;
-  ksba_cert_t issuer_cert = NULL;
-  FILE *fp;
+  crl_cache_result_t result;
+  unsigned char issuerhash[20];
+  char issuerhash_hex[41];
+  ksba_sexp_t serial;
+  unsigned char *sn;
+  size_t snlen;
+  char *endp, *tmp;
+  int i;
 
+  /* Compute the hash value of the issuer name.  */
+  tmp = ksba_cert_get_issuer (cert, 0);
+  if (!tmp)
+    {
+      log_error ("oops: issuer missing in certificate\n");
+      return gpg_error (GPG_ERR_INV_CERT_OBJ);
+    }
+  gcry_md_hash_buffer (GCRY_MD_SHA1, issuerhash, tmp, strlen (tmp));
+  xfree (tmp);
+  for (i=0,tmp=issuerhash_hex; i < 20; i++, tmp += 2)
+    sprintf (tmp, "%02X", issuerhash[i]);
+  
+  /* Get the serial number.  */
+  serial = ksba_cert_get_serial (cert);
+  if (!serial)
+    {
+      log_error ("oops: S/N missing in certificate\n");
+      return gpg_error (GPG_ERR_INV_CERT_OBJ);
+    }
+  sn = serial;
+  if (*sn != '(')
+    {
+      log_error ("oops: invalid S/N\n");
+      xfree (serial);
+      return gpg_error (GPG_ERR_INV_CERT_OBJ);
+    }
+  sn++;
+  snlen = strtoul (sn, &endp, 10);
+  sn = endp;
+  if (*sn != ':')
+    {
+      log_error ("oops: invalid S/N\n");
+      xfree (serial);
+      return gpg_error (GPG_ERR_INV_CERT_OBJ);
+    }
+  sn++;
+
+  /* Check the cache.  */
+  result = cache_isvalid (ctrl, issuerhash_hex, sn, snlen, force_refresh);
+  switch (result)
+    {
+    case CRL_CACHE_VALID:
+      err = 0;
+      break;
+    case CRL_CACHE_INVALID:
+      err = gpg_error (GPG_ERR_CERT_REVOKED);
+      break;
+    case CRL_CACHE_DONTKNOW: 
+      err = gpg_error (GPG_ERR_NO_CRL_KNOWN);
+    case CRL_CACHE_CANTUSE: 
+      err = gpg_error (GPG_ERR_NO_CRL_KNOWN);
+      break;
+    default:
+      log_fatal ("cache_isvalid returned invalid status code %d\n", result);
+    }
+
+  xfree (serial);
+  return err;
+}  
+
+
+/* Return the certificate for NAME or NULL if it can't be found.  Note
+   that we will only return the first matching certificate. */
+static ksba_cert_t 
+get_crlissuer_cert (ctrl_t ctrl, const char *name)
+{
+  gpg_error_t err;
+  ksba_cert_t cert = NULL;
+  cert_fetch_context_t context;
+
+  /* First check whether it has already been cached.  */
+  cert = get_cert_bysubject (name, 0); 
+  if (cert)
+    return cert; /* Done. */
+    
   /* Ask back to the service requester to return the certificate.
      This is because we can assume that he already used the
      certificate while checking for the CRL. */
-  issuer_cert = get_cert_local (ctrl, issuer);
-  if (issuer_cert)
-    return issuer_cert; /* Done. */
+  cert = get_cert_local (ctrl, name);
+  if (cert)
+    return cert; /* Done. */
 
   /* If this fails do a regular certificate lookup. */
   if (DBG_LOOKUP)
     log_debug ("certificate for CRL issuer not returned by caller"
                " - doing lookup\n");
 
-  err = ca_cert_fetch (issuer, &fp);
+  err = ca_cert_fetch (ctrl, &context, name);
   if (err)
     {
-      log_error (_("error fetching certificate for issuer: %s\n"),
+      log_error (_("error fetching certificate for CRL issuer: %s\n"),
                  gpg_strerror (err));
       return NULL;
     }
 
-  err = ksba_cert_new (&issuer_cert);
-  if (!err)
-    {
-      ksba_reader_t r;
-
-      err = ksba_reader_new (&r);
-      if (!err)
-        err = ksba_reader_set_file (r, fp);
-      if (!err)
-        err = ksba_cert_read_der (issuer_cert, r);
-      if (err)
-        {
-          ksba_cert_release (issuer_cert);
-          issuer_cert = NULL;
-        }
-      ksba_reader_release (r);
-    }
+  err = fetch_next_ksba_cert (context, &cert);
   if (err)
-    log_error (_("invalid issuer certificate: %s\n"), gpg_strerror (err) );
-  fclose (fp);
+    log_error (_("invalid CRL issuer certificate: %s\n"), gpg_strerror (err) );
+  end_cert_fetch (context);
 
-  return issuer_cert;
+  if (!err)
+    cache_cert (cert);
+
+  return cert;
+}
+
+
+/* Return the certificate from the issuer DN NAME and the S/N SN.
+   First tries the local chache, then asks the caller and then tries
+   an ldap search. */
+static ksba_cert_t 
+get_crlissuer_cert_bysn (ctrl_t ctrl, const char *name, ksba_sexp_t sn)
+{
+  ksba_cert_t cert;
+  char *snbuf, *buf;
+
+  /* First check whether it has already been cached.  */
+  cert = get_cert_bysn (name, sn);
+  if (cert)
+    return cert; /* Done. */
+
+  /* Ask back to the service requester to return the certificate.
+     This is because we can assume that he already used the
+     certificate while checking for the CRL. */
+  snbuf = serial_hex (sn);
+  if (!snbuf)
+    snbuf = "";
+  buf = xtrymalloc (1 + strlen (snbuf) + 1 + strlen (name) + 1 );
+  if (!buf)
+    {
+      xfree (snbuf);
+      return NULL;
+    }
+  strcpy (stpcpy (stpcpy (stpcpy (buf, "#"), snbuf), "/"), name);
+  xfree (snbuf);
+  cert = get_cert_local (ctrl, buf);
+  xfree (buf);
+  if (cert)
+    {
+      cache_cert (cert);
+      return cert; /* Done. */
+    }
+
+  return find_cert_bysn (ctrl, name, sn);
 }
 
 
@@ -1303,7 +1475,9 @@ start_sig_check (ksba_crl_t crl, gcry_md_hd_t *md, int *algo)
                  *algo, gcry_strerror (err));
       return err;
     }
-  
+  if (DBG_HASHING)
+    gcry_md_start_debug (*md, "crl");
+
   ksba_crl_set_hash_function (crl, HASH_FNC, *md);  
   return 0;
 }
@@ -1346,6 +1520,8 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
     }
 	
   /* Get and convert the public key for the issuer certificate. */
+  if (DBG_X509)
+    dump_cert ("crl_issuer_cert", issuer_cert);
   pubkey = ksba_cert_get_public_key (issuer_cert);
   n = gcry_sexp_canon_len (pubkey, 0, NULL, NULL);
   if (!n) 
@@ -1378,6 +1554,8 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
 
   /* Pass this on to the signature verification. */
   err = gcry_pk_verify (s_sig, s_hash, s_pkey);
+  if (DBG_X509)
+    log_debug ("gcry_pk_verify: %s\n", gpg_strerror (err));
 
  leave:
   xfree (sigval);
@@ -1394,28 +1572,29 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
    CRL object and stored in the data base file DB with the name FNAME
    (only used for printing error messages).  That DB should be a
    temporary one and not the actual one.  If the function fails the
-   caller should delete this temporary databse file.  CTRL is required
-   to retrieve certificates using the general dirmngr callback
-   service.  R_ISSUER returns an allocated string with the issuer DB,
-   THIS_UPDATE and NEXT_UPDATE are filled with the correspondind data
-   from the CRL.  Note that these values might get set even if the CRL
-   processing fails at a later step; thus the caller should free
-   *R_ISSUER even when the function returns with an error.
+   caller should delete this temporary database file.  CTRL is
+   required to retrieve certificates using the general dirmngr
+   callback service.  R_CRLISSUER returns an allocated string with the
+   crl-issuer DN, THIS_UPDATE and NEXT_UPDATE are filled with the
+   correspondind data from the CRL.  Note that these values might get
+   set even if the CRL processing fails at a later step; thus the
+   caller should free *R_ISSUER even if the function returns with an
+   error.
 */ 
 static int 
 crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
                   struct cdb_make *cdb, const char *fname,
-                  char **r_issuer,
+                  char **r_crlissuer,
                   ksba_isotime_t thisupdate, ksba_isotime_t nextupdate)
 {  
   gpg_error_t err;
   ksba_stop_reason_t stopreason;
-  ksba_cert_t issuer_cert = NULL;
+  ksba_cert_t crlissuer_cert = NULL;
   gcry_md_hd_t md = NULL;
   int algo = 0;
   size_t n;
 
-  *r_issuer = NULL;
+  *r_crlissuer = NULL;
   *thisupdate = *nextupdate = 0;
 
   /* Start of the KSBA parser loop. */
@@ -1432,29 +1611,9 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
         {
         case KSBA_SR_BEGIN_ITEMS:
           {
-            char *issuer;
-
             if (start_sig_check (crl, &md, &algo ))
               goto failure;
           
-            err = ksba_crl_get_issuer (crl, &issuer);
-            if( err )
-              {
-                log_error (_("no issuer found in CRL: %s\n"),
-                           gpg_strerror (err) );
-                err = gpg_error (GPG_ERR_INV_CRL);
-                goto failure;
-              }
-            *r_issuer = issuer; /* (Do it here so we don't need to
-                                    free it later) */
-
-            issuer_cert = get_issuer_cert (ctrl, issuer);
-            if (!issuer_cert)
-              {
-                err = gpg_error (GPG_ERR_MISSING_CERT);
-                goto failure;
-              }
-        
             err = ksba_crl_get_update_times (crl, thisupdate, nextupdate);
             if (err)
               {
@@ -1467,8 +1626,6 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
             if (opt.verbose)
               log_info (_("update times of this CRL: this=%s next=%s\n"), 
                         thisupdate, nextupdate);   
-
-            issuer = NULL;
           }
           break;
       
@@ -1514,13 +1671,73 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
       
         case KSBA_SR_READY:
           {
-            err = finish_sig_check (crl, md, algo, issuer_cert);
+            char *crlissuer;
+            ksba_name_t authid;
+            ksba_sexp_t authidsn;
+
+            /* We need to look for the issuer only after having read
+               all items.  The issuer itselfs comes before the items
+               but the optional authorityKeyIdentifier comes after the
+               items. */
+            err = ksba_crl_get_issuer (crl, &crlissuer);
+            if( err )
+              {
+                log_error (_("no CRL issuer found in CRL: %s\n"),
+                           gpg_strerror (err) );
+                err = gpg_error (GPG_ERR_INV_CRL);
+                goto failure;
+              }
+            *r_crlissuer = crlissuer; /* (Do it here so we don't need
+                                         to free it later) */
+
+            if (!ksba_crl_get_auth_key_id (crl, NULL, &authid, &authidsn))
+              {
+                const char *s;
+
+                if (opt.verbose)
+                  log_info (_("locating CRL issuer certificate by "
+                              "authorityKeyIdentifier\n"));
+                
+                s = ksba_name_enum (authid, 0);
+                if (s && *authidsn)
+                  crlissuer_cert = get_crlissuer_cert_bysn (ctrl, s, authidsn);
+                if (!crlissuer_cert)
+                  {
+                    log_info ("CRL issuer certificate (#");
+                    dump_serial (authidsn);
+                    log_printf ("/");
+                    dump_string (s);
+                    log_printf (") not found\n");
+                  }
+                ksba_name_release (authid);
+                xfree (authidsn);
+              }
+            else
+              crlissuer_cert = get_crlissuer_cert (ctrl, crlissuer);
+            err = 0;
+            if (!crlissuer_cert)
+              {
+                err = gpg_error (GPG_ERR_MISSING_CERT);
+                goto failure;
+              }
+        
+            err = finish_sig_check (crl, md, algo, crlissuer_cert);
             if (err)
               {
                 log_error (_("CRL signature verification failed: %s\n"), 
                            gpg_strerror (err));
                 goto failure;
               }
+            err = validate_cert_chain (ctrl, crlissuer_cert,
+                                       NULL, VALIDATE_MODE_CRL);
+            if (err)
+              {
+                log_error (_("error checking validity of CRL "
+                             "issuer certificate: %s\n"), 
+                           gpg_strerror (err));
+                goto failure;
+              }
+
           }
           break;
       
@@ -1535,7 +1752,7 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
 
 
  failure:
-  ksba_cert_release (issuer_cert);
+  ksba_cert_release (crlissuer_cert);
   return err;
 }
 
@@ -1560,7 +1777,7 @@ get_crl_number (ksba_crl_t crl)
 
 
 /* Return the authorityKeyIdentifier or NULL if it is not available.
-   The issuer name may consists of several parts - the are delimted by
+   The issuer name may consists of several parts - they are delimted by
    0x01. */
 static char *
 get_auth_key_id (ksba_crl_t crl, char **serialno)
@@ -1610,11 +1827,11 @@ get_auth_key_id (ksba_crl_t crl, char **serialno)
    CACHE.  The CRL itself will be read from the stream FP and is
    expected in binary format. */
 gpg_error_t
-crl_cache_insert (ctrl_t ctrl, crl_cache_t cache, const char *url, FILE *fp)
+crl_cache_insert (ctrl_t ctrl, const char *url, ksba_reader_t reader)
 {
+  crl_cache_t cache = get_current_cache ();
   gpg_error_t err, err2;
   ksba_crl_t crl;
-  ksba_reader_t reader;
   char *fname = NULL;
   char *newfname = NULL;
   struct cdb_make cdb;
@@ -1630,6 +1847,11 @@ crl_cache_insert (ctrl_t ctrl, crl_cache_t cache, const char *url, FILE *fp)
   int idx;
   const char *oid;
   int critical;
+
+
+  /* FIXME: We should acquire a mutex for the URL, so that we don't
+     simultaneously enter the same CRL twice.  However this needs to be
+     interweaved with the checking function.*/
  
   err2 = 0;
 
@@ -1640,14 +1862,6 @@ crl_cache_insert (ctrl_t ctrl, crl_cache_t cache, const char *url, FILE *fp)
       goto leave;
     }
   
-  err = ksba_reader_new (&reader);
-  if (!err)
-    err = ksba_reader_set_file (reader, fp);
-  if (err)
-    {
-      log_error (_("ksba_reader_set_file failed: %s\n"), gpg_strerror (err));
-      goto leave;
-    }
   err = ksba_crl_set_reader (crl, reader);
   if ( err )
     {
@@ -1671,7 +1885,7 @@ crl_cache_insert (ctrl_t ctrl, crl_cache_t cache, const char *url, FILE *fp)
     for (p=tmpfname; *p; p++)
       if (*p == '/')
         *p = '.';
-    fname = make_filename (opt.homedir, DBDIR_D, tmpfname, NULL);
+    fname = make_filename (opt.homedir_cache, DBDIR_D, tmpfname, NULL);
     xfree (tmpfname);
     if (!remove (fname))
       log_info (_("removed stale temporary cache file `%s'\n"), fname);
@@ -1758,7 +1972,7 @@ crl_cache_insert (ctrl_t ctrl, crl_cache_t cache, const char *url, FILE *fp)
           || !strcmp (oid, oidstr_authorityKeyIdentifier)
           || !strcmp (oid, oidstr_crlNumber) )
         continue;
-      log_error (_("unknown crtical CRL extension %s\n"), oid);
+      log_error (_("unknown critical CRL extension %s\n"), oid);
       if (!err2)
         err2 = gpg_error (GPG_ERR_INV_CRL);
       invalidate_crl |= 2;
@@ -1836,7 +2050,6 @@ crl_cache_insert (ctrl_t ctrl, crl_cache_t cache, const char *url, FILE *fp)
       xfree (fname);
     }
   xfree (newfname);
-  ksba_reader_release (reader);
   ksba_crl_release (crl);
   xfree (issuer);
   xfree (issuer_hash);
@@ -1967,8 +2180,9 @@ list_one_crl_entry (crl_cache_t cache, crl_cache_entry_t e, FILE *fp)
 /* Print the contents of the CRL CACHE in a human readable format to
    stream FP. */
 gpg_error_t 
-crl_cache_list (crl_cache_t cache, FILE *fp) 
+crl_cache_list (FILE *fp) 
 {
+  crl_cache_t cache = get_current_cache ();
   crl_cache_entry_t entry;
   gpg_error_t err = 0;
 
@@ -1981,13 +2195,13 @@ crl_cache_list (crl_cache_t cache, FILE *fp)
 }
 
 
-/* Load the CRL contain the the file named FILENAME into our cache
-   CACHE */
+/* Load the CRL containing the file named FILENAME into our CRL cache. */
 gpg_error_t
-crl_cache_load (ctrl_t ctrl, crl_cache_t cache, const char *filename)
+crl_cache_load (ctrl_t ctrl, const char *filename)
 {
   gpg_error_t err;
-  FILE* fp;
+  FILE *fp;
+  ksba_reader_t reader;
 
   fp = fopen (filename, "r");
   if (!fp)
@@ -1997,8 +2211,155 @@ crl_cache_load (ctrl_t ctrl, crl_cache_t cache, const char *filename)
       return err;
     }
 
-  err = crl_cache_insert (ctrl, cache, filename, fp);
+  err = ksba_reader_new (&reader);
+  if (!err)
+    err = ksba_reader_set_file (reader, fp);
+  if (err)
+    {
+      log_error (_("error initializing reader object: %s\n"),
+                 gpg_strerror (err));
+      ksba_reader_release (reader);
+      return err;
+    }
+  err = crl_cache_insert (ctrl, filename, reader);
+  ksba_reader_release (reader);
   fclose (fp);
   return err;
 }
 
+
+/* Locate the corresponding CRL for the certificate CERT, read and
+   verify the CRL and store it in the cache.  */
+gpg_error_t
+crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert) 
+{
+  gpg_error_t err;
+  ksba_reader_t reader = NULL;
+  char *issuer = NULL;
+  ksba_name_t distpoint = NULL;
+  ksba_name_t issuername = NULL;
+  char *distpoint_uri = NULL;
+  char *issuername_uri = NULL;
+  int any_dist_point = 0;
+  int seq;
+
+  /* Loop over all distribution points, get the CRLs and put them into
+     the cache. */
+  if (DBG_X509)
+    log_debug ("checking distribution points\n");
+  seq = 0;
+  while ( !(err = ksba_cert_get_crl_dist_point (cert, seq++,
+                                                &distpoint,
+                                                &issuername, NULL )))
+    {
+      if (!distpoint && !issuername)
+        {
+          if (DBG_X509)
+            log_debug ("no issuer name and no distribution point\n");
+          break; /* Not allowed; i.e. an invalid certificate.  We give
+                    up here and hope that the default method returns a
+                    suitable CRL. */
+        }
+
+      /* Get the URIs. */
+      xfree (distpoint_uri); distpoint_uri = NULL;
+      xfree (issuername_uri); issuername_uri = NULL;
+      distpoint_uri = ksba_name_get_uri (distpoint, 0); 
+      issuername_uri = ksba_name_get_uri (issuername, 0); 
+      ksba_name_release (distpoint); distpoint = NULL;
+      ksba_name_release (issuername); issuername = NULL;
+
+      if (!distpoint_uri)
+        continue;
+
+      if (!strncmp (distpoint_uri, "ldap:", 5)
+          || !strncmp (distpoint_uri, "ldaps:", 6))
+        {
+          if (opt.ignore_ldap_dp)
+            continue;
+        }
+      else if (!strncmp (distpoint_uri, "http:", 5)
+               || !strncmp (distpoint_uri, "https:", 6))
+        {
+          if (opt.ignore_http_dp)
+            continue;
+        }
+      else
+        continue; /* Skip unknown schemes. */
+
+      any_dist_point = 1;
+      
+      if (DBG_X509)
+        log_debug ("fetching CRL from `%s'\n", distpoint_uri);
+      err = crl_fetch (ctrl, distpoint_uri, &reader);
+      if (err)
+        {
+          log_error (_("crl_fetch via DP failed: %s\n"), gpg_strerror (err));
+          goto leave;
+        }
+      
+      if (DBG_X509)
+        log_debug ("inserting CRL\n");
+      err = crl_cache_insert (ctrl, distpoint_uri, reader); 
+      if (err)
+        {
+          log_error (_("crl_cache_insert via DP failed: %s\n"),
+                     gpg_strerror (err));
+          goto leave;
+        }
+    }
+  if (gpg_err_code (err) == GPG_ERR_EOF)
+    err = 0;
+
+  /* If we did not found any distpoint, try something reasonable. */
+  if (!any_dist_point )
+    {
+      if (DBG_X509)
+        log_debug ("no distribution point - trying issuer name\n");
+      
+      if (reader)
+        {
+          crl_close_reader (reader);
+          reader = NULL;
+        }
+
+      issuer = ksba_cert_get_issuer (cert, 0);
+      if (!issuer) 
+        {
+          log_error ("oops: issuer missing in certificate\n");
+          err = gpg_error (GPG_ERR_INV_CERT_OBJ); 
+          goto leave;
+        }
+
+      if (DBG_X509)
+        log_debug ("fetching CRL from default location\n");
+      err = crl_fetch_default (ctrl, issuer, &reader);
+      if (err)
+          {
+            log_error (_("crl_fetch via issuer failed: %s\n"),
+                       gpg_strerror (err));
+            goto leave;
+          }
+
+      if (DBG_X509)
+        log_debug ("inserting CRL\n");
+      err = crl_cache_insert (ctrl, "default location(s)", reader);
+      if (err)
+        {
+          log_error (_("crl_cache_insert via issuer failed: %s\n"),
+                     gpg_strerror (err));
+          goto leave;
+        }
+    }
+
+ leave:
+  if (reader)
+    crl_close_reader (reader);
+  xfree (distpoint_uri);
+  xfree (issuername_uri);
+  ksba_name_release (distpoint); 
+  ksba_name_release (issuername); 
+  ksba_free (issuer);
+  return err;
+}
+
