@@ -1,5 +1,5 @@
 /* http.c  -  HTTP protocol handler
- *      Copyright (C) 1999, 2001, 2002, 2003 Free Software Foundation, Inc.
+ * Copyright (C) 1999, 2001, 2002, 2003, 2004 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -63,7 +63,7 @@ static int remove_escapes (unsigned char *string);
 static int insert_escapes (unsigned char *buffer, const unsigned char *string,
                            const unsigned char *special);
 static uri_tuple_t parse_tuple (unsigned char *string);
-static gpg_error_t send_request (http_t hd);
+static gpg_error_t send_request (http_t hd, const char *proxy);
 static unsigned char *build_rel_path (parsed_uri_t uri);
 static gpg_error_t parse_response (http_t hd);
 
@@ -72,7 +72,8 @@ static gpg_error_t write_server (int sock, const char *data, size_t length);
 
 
 gpg_error_t
-http_open (http_t hd, http_req_t reqtype, const char *url, unsigned int flags)
+http_open (http_t hd, http_req_t reqtype, const char *url, unsigned int flags,
+           const char *proxy)
 {
   gpg_error_t err;
 
@@ -89,7 +90,7 @@ http_open (http_t hd, http_req_t reqtype, const char *url, unsigned int flags)
   err = http_parse_uri (&hd->uri, url);
   if (!err)
     {
-      err = send_request (hd);
+      err = send_request (hd, proxy);
       if (!err)
         {
           hd->fp_write = fdopen (hd->sock, "w");
@@ -156,13 +157,16 @@ http_wait_response (http_t hd, unsigned int *ret_status)
 
 
 /* Convenience function to send a request and wait for the response.
-   Closes the handle on error. */
+   Closes the handle on error.  If PROXY is not NULL, this value will
+   be used as a HTTP proxy and any enabled $http_proxy gets
+   ignored. */
 gpg_error_t
-http_open_document (http_t hd, const char *document, unsigned int flags)
+http_open_document (http_t hd, const char *document, unsigned int flags,
+                    const char *proxy)
 {
   gpg_error_t err;
 
-  err = http_open (hd, HTTP_REQ_GET, document, flags);
+  err = http_open (hd, HTTP_REQ_GET, document, flags, proxy);
   if (err)
     return err;
 
@@ -392,7 +396,7 @@ insert_escapes (unsigned char *buffer, const unsigned char *string,
         {
           if (buffer)
             {
-              sprintf (buffer, "%02X", *string);
+              sprintf (buffer, "%%%02X", *string);
               buffer += 3;
             }
           n += 3;
@@ -444,7 +448,7 @@ parse_tuple (unsigned char *string)
  * Returns 0 if the request was successful
  */
 static gpg_error_t
-send_request (http_t hd)
+send_request (http_t hd, const char *proxy)
 {
   const unsigned char *server;
   unsigned char *request, *p;
@@ -456,17 +460,20 @@ send_request (http_t hd)
   server = *hd->uri->host ? hd->uri->host : "localhost";
   port = hd->uri->port ? hd->uri->port : 80;
 
-  if ((hd->flags & HTTP_FLAG_TRY_PROXY)
-      && (http_proxy = getenv (HTTP_PROXY_ENV)))
+  if (proxy
+      || ( (hd->flags & HTTP_FLAG_TRY_PROXY)
+           && (http_proxy = getenv (HTTP_PROXY_ENV))))
     {
       parsed_uri_t uri;
+      
+      if (proxy)
+        http_proxy = proxy;
 
       err = http_parse_uri (&uri, http_proxy);
       if (err)
         {
-          log_error (_("invalid %s: %s\n"),
-                     HTTP_PROXY_ENV_PRINTABLE,
-                     gpg_strerror (err));
+          log_error (_("invalid HTTP proxy (%s): %s\n"),
+                     http_proxy, gpg_strerror (err));
           http_release_parsed_uri (uri);
           return gpg_error (GPG_ERR_CONFIGURATION);
         }
