@@ -1,5 +1,5 @@
 /* misc.h - miscellaneous
- *      Copyright (C) 2002 Klarlvdalens Datakonsult AB
+ *      Copyright (C) 2002 Klarälvdalens Datakonsult AB
  *
  * This file is part of DirMngr.
  *
@@ -84,6 +84,29 @@ gpg_error_t canon_sexp_to_gcry (const unsigned char *canon,
 /* Return an allocated hex-string with the SHA-1 fingerprint of
    CERT. */
 char *get_fingerprint_hexstring (ksba_cert_t cert);
+/* Return an allocated hex-string with the SHA-1 fingerprint of
+   CERT.  This version inserts the usual colons. */
+char *get_fingerprint_hexstring_colon (ksba_cert_t cert);
+
+
+/* Dump the serial number SERIALNO to the log stream.  */
+void dump_serial (ksba_sexp_t serialno);
+
+/* Dump the ISO time T to the log stream without a LF.  */
+void dump_isotime (dirmngr_isotime_t t);
+
+/* Dump STRING to the log file but choose the best readable
+   format.  */
+void dump_string (const char *string);
+
+/* Dump an KSBA cert object to the log stream. Prefix the output with
+   TEXT.  This is used for debugging. */
+void dump_cert (const char *text, ksba_cert_t cert);
+
+/* Return the host name and the port (0 if none was given) from the
+   URL.  Return NULL on error or if host is not included in the
+   URL.  */
+char *host_and_port_from_url (const char *url, int *port);
 
 
 #ifdef HAVE_FOPENCOOKIE
