@@ -21,7 +21,6 @@
 #ifndef OCSP_H
 #define OCSP_H
 
-gpg_error_t ocsp_isvalid (ctrl_t ctrl, const char *cert_fpr);
-
+gpg_error_t ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr);
 
 #endif /*OCSP_H*/
