@@ -1,5 +1,5 @@
 /* dirmngr.c - LDAP access
- *	Copyright (C) 2002 Klarlvdalens Datakonsult AB
+ *	Copyright (C) 2002 Klarälvdalens Datakonsult AB
  *      Copyright (C) 2003, 2004 g10 Code GmbH
  *
  * This file is part of DirMngr.
@@ -37,68 +37,22 @@
 #include "crlfetch.h"
 #include "dirmngr.h"
 #include "ocsp.h"
+#include "certcache.h"
+#include "validate.h"
+
+/* To avoid DoS attacks we limit the size of a certificate to
+   something reasonable. */
+#define MAX_CERT_LENGTH (8*1024)
 
 #define PARM_ERROR(t) assuan_set_error (ctx, ASSUAN_Parameter_Error, (t))
 
 
-/* We keep track of the primary client using us.  This is mainly a
-   hack to send the progress info back. */
-static ctrl_t primary_connection;
-
 /* Data used to associate an Assuan context with local server data */
 struct server_local_s {
   ASSUAN_CONTEXT assuan_ctx;
-  crl_cache_t     crl_cache;
-  int message_fd;
 };
 
 
-/* Map Assuan error code ERR to an GPG_ERR_ code.  We need to
-   distinguish between genuine (and legacy) Assuan error codes and
-   application error codes shared with all GnuPG modules.  The rule is
-   simple: All errors with a gpg_err_source of UNKNOWN are genuine;
-   all other Assuan codes are passed verbatim through. */
-static gpg_error_t
-map_assuan_err (int err)
-{
-  gpg_err_code_t ec;
-
-  if (gpg_err_source (err))
-    return err;
-
-  switch (err)
-    {
-    case -1:                     ec = GPG_ERR_EOF; break;
-    case 0:                      ec = 0; break;
-
-    case ASSUAN_Canceled:        ec = GPG_ERR_CANCELED; break;
-    case ASSUAN_Invalid_Index:   ec = GPG_ERR_INV_INDEX; break;
-
-    case ASSUAN_Not_Implemented: ec = GPG_ERR_NOT_IMPLEMENTED; break;
-    case ASSUAN_Server_Fault:    ec = GPG_ERR_ASSUAN_SERVER_FAULT; break;
-    case ASSUAN_No_Public_Key:   ec = GPG_ERR_NO_PUBKEY; break;
-    case ASSUAN_No_Secret_Key:   ec = GPG_ERR_NO_SECKEY; break;
-
-    case ASSUAN_Cert_Revoked:    ec = GPG_ERR_CERT_REVOKED; break;
-    case ASSUAN_No_CRL_For_Cert: ec = GPG_ERR_NO_CRL_KNOWN; break;       
-    case ASSUAN_CRL_Too_Old:     ec = GPG_ERR_CRL_TOO_OLD; break;        
-
-    case ASSUAN_Not_Trusted:     ec = GPG_ERR_NOT_TRUSTED; break;
-
-    case ASSUAN_Card_Error:      ec = GPG_ERR_CARD; break;
-    case ASSUAN_Invalid_Card:    ec = GPG_ERR_INV_CARD; break;
-    case ASSUAN_No_PKCS15_App:   ec = GPG_ERR_NO_PKCS15_APP; break;
-    case ASSUAN_Card_Not_Present: ec= GPG_ERR_CARD_NOT_PRESENT; break;
-    case ASSUAN_Not_Confirmed:   ec = GPG_ERR_NOT_CONFIRMED; break;
-    case ASSUAN_Invalid_Id:      ec = GPG_ERR_INV_ID; break;
-
-    default:
-      ec = err < 100? GPG_ERR_ASSUAN_SERVER_FAULT : GPG_ERR_ASSUAN;
-      break;
-    }
-  return gpg_err_make (GPG_ERR_SOURCE_UNKNOWN, ec);
-}
-
 /* Map GPG_ERR_xx error codes to Assuan status codes */
 static int
 map_to_assuan_status (int rc)
@@ -129,8 +83,8 @@ map_to_assuan_status (int rc)
    target string D as long as the source string S, i.e.: strlen(s)+1.
    NOte further that If S contains an escaped binary nul the resulting
    string D will contain the 0 as well as all other characters but it
-   will be impossible to kno whether this is the original eos or a
-   copied nul. */
+   will be impossible to know whether this is the original EOS or a
+   copied Nul. */
 static void
 strcpy_escaped_plus (char *d, const unsigned char *s)
 {
@@ -169,14 +123,8 @@ do_get_cert_local (ctrl_t ctrl, const char *name, const char *command)
   else
     buf = xstrdup (command);
 
-  /* FIXME: When changed to run as a system wide server, we should
-     setup a maxlen limit for the certificate.  The problem is that
-     libassuan allocates a buffer of that size in advance which we
-     usually don't need. As long as dirmngr is used in server mode we
-     don't need to care about it because the user can toast his
-     applications in many other ways.*/
   rc = assuan_inquire (ctrl->server_local->assuan_ctx, buf,
-                       &value, &valuelen, 0);
+                       &value, &valuelen, MAX_CERT_LENGTH);
   xfree (buf);
   if (rc)
     {
@@ -184,6 +132,12 @@ do_get_cert_local (ctrl_t ctrl, const char *name, const char *command)
       return NULL;
     }
   
+  if (!valuelen)
+    {
+      xfree (value);
+      return NULL;
+    }
+
   rc = ksba_cert_new (&cert);
   if (!rc)
     {
@@ -236,6 +190,8 @@ get_issuing_cert_local (ctrl_t ctrl, const char *name)
 }
 
 
+
+
 /* Ask the client to return the certificate asscociated with the
    current command. This is sometimes needed because the client usually
    sends us just the cert ID, assuming that the request can be
@@ -246,17 +202,9 @@ inquire_cert_and_load_crl (assuan_context_t ctx)
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err;
   assuan_error_t ae;
-  FILE *crl_fp = NULL;
   unsigned char *value = NULL;
   size_t valuelen; 
-  char *issuer = NULL;
-  ksba_cert_t issuer_cert = NULL;
-  int seq;
-  int found_dist_point = 0;
-  ksba_name_t distpoint = NULL;
-  ksba_name_t issuername = NULL;
-  char *distpoint_uri = NULL;
-  char *issuername_uri = NULL;
+  ksba_cert_t cert = NULL;
 
   ae = assuan_inquire( ctx, "SENDCERT", &value, &valuelen, 0);
   if (ae)
@@ -272,104 +220,18 @@ inquire_cert_and_load_crl (assuan_context_t ctx)
   if (!valuelen) /* No data returned; return a comprehensible error. */
     return gpg_error (GPG_ERR_MISSING_CERT);
 
-  err = ksba_cert_new (&issuer_cert);
+  err = ksba_cert_new (&cert);
   if (err)
     goto leave;
-  err = ksba_cert_init_from_mem (issuer_cert, value, valuelen);
+  err = ksba_cert_init_from_mem (cert, value, valuelen);
   if(err)
     goto leave;
+  xfree (value); value = NULL;
 
-  xfree (value); 
-  value = NULL;
-
-  issuer = ksba_cert_get_issuer (issuer_cert, 0);
-  if (!issuer) 
-    return gpg_error (GPG_ERR_BAD_CERT); /* No issuer in certificate. */
-
-  /* Loop over all distribution points, get the CRLs and put them
-     into the cache. */
-  if (DBG_X509)
-    log_debug ("checking distribution points\n");
-  seq = 0;
-  while ( !(err = ksba_cert_get_crl_dist_point (issuer_cert, seq++,
-                                                &distpoint,
-                                                &issuername, NULL )))
-    {
-      if (!distpoint && !issuername)
-        {
-          if (DBG_X509)
-            log_debug ("no issuername and no distpoint\n");
-          break; /* Not allowed, so don't keep trying. */
-        }
-
-      /* Get the URIs. */
-      xfree (distpoint_uri); distpoint_uri = NULL;
-      xfree (issuername_uri); issuername_uri = NULL;
-      distpoint_uri = ksba_name_get_uri (distpoint, 0); 
-      issuername_uri = ksba_name_get_uri (issuername, 0); 
-      ksba_name_release (distpoint); distpoint = NULL;
-      ksba_name_release (issuername); issuername = NULL;
-      found_dist_point = 1;
-      
-      xfree (value); value = NULL;
-      err = crl_fetch (distpoint_uri, &crl_fp);
-      if (err)
-        {
-          log_error (_("crl_fetch via DP failed: %s\n"), gpg_strerror (err));
-          goto leave;
-        }
-      
-      err = crl_cache_insert (ctrl, ctrl->server_local->crl_cache,
-                              distpoint_uri, crl_fp); 
-      if (err)
-        {
-          log_error (_("crl_cache_insert via DP failed: %s\n"),
-                     gpg_strerror (err));
-          goto leave;
-        }
-    }
-  if (gpg_err_code (err) == GPG_ERR_EOF)
-    err = 0;
-
-  /* If we did not found any distpoint, try something reasonable. */
-  if (!found_dist_point )
-    {
-      if (DBG_X509)
-        log_debug ("no distribution point - trying issuer name\n");
-      
-      if (crl_fp)
-        {
-          fclose (crl_fp);
-          crl_fp = NULL;
-        }
-      xfree (value); value = NULL;
-      err = crl_fetch_default (issuer, &crl_fp);
-      if (err)
-          {
-            log_error (_("crl_fetch via issuer failed: %s\n"),
-                       gpg_strerror (err));
-            goto leave;
-          }
-
-      err = crl_cache_insert (ctrl, ctrl->server_local->crl_cache,
-                             "default location(s)", crl_fp);
-      if (err)
-        {
-          log_error (_("crl_cache_insert via issuer failed: %s\n"),
-                     gpg_strerror (err));
-          goto leave;
-        }
-    }
+  err = crl_cache_reload_crl (ctrl, cert);
 
  leave:
-  if (crl_fp)
-    fclose (crl_fp);
-  xfree (distpoint_uri);
-  xfree (issuername_uri);
-  ksba_name_release (distpoint); 
-  ksba_name_release (issuername); 
-  ksba_free (issuer);
-  ksba_cert_release (issuer_cert);
+  ksba_cert_release (cert);
   xfree (value);
   return err;
 }
@@ -451,7 +313,7 @@ cmd_isvalid (assuan_context_t ctx, char *line)
       if (!opt.allow_ocsp)
         err = gpg_error (GPG_ERR_NOT_SUPPORTED);
       else
-        err = ocsp_isvalid (ctrl, NULL);
+        err = ocsp_isvalid (ctrl, NULL, NULL);
       /* Fixme: If we got no ocsp response we should fall back to CRL
          mode.  Thus we need to clear OCSP_MODE, get the issuerhash
          and the serialno from the current certificate and jump to
@@ -459,7 +321,7 @@ cmd_isvalid (assuan_context_t ctx, char *line)
     }
   else
     {
-      switch (crl_cache_isvalid (ctrl->server_local->crl_cache,
+      switch (crl_cache_isvalid (ctrl,
                                  issuerhash, serialno,
                                  ctrl->force_crl_refresh))
         {
@@ -493,6 +355,201 @@ cmd_isvalid (assuan_context_t ctx, char *line)
 }
 
 
+/* If the line contains a SHA-1 fingerprint as the first argument,
+   return the FPR vuffer on success.  The function checks that the
+   fingerprint consists of valid characters and prints and error
+   message if it does not and returns NULL.  Fingerprints are
+   considered optional and thus no explicit error is returned. NULL is
+   also returned if there is no fingerprint at all available. 
+   FPR must be a caller provided buffer of at least 20 bytes.
+
+   Note that colons within the fingerprint are allowed to separate 2
+   hex digits; this allows for easier cutting and pasting using the
+   usual fingerprint rendering.
+*/
+static unsigned char *
+get_fingerprint_from_line (const char *line, unsigned char *fpr)
+{
+  const char *s;
+  int i;
+
+  for (s=line, i=0; *s && *s != ' '; s++ )
+    {
+      if ( hexdigitp (s) && hexdigitp (s+1) )
+        {
+          if ( i >= 20 )
+            return NULL;  /* Fingerprint too long.  */
+          fpr[i++] = xtoi_2 (s);
+          s++;
+        }
+      else if ( *s != ':' )
+        return NULL; /* Invalid.  */
+    }
+  if ( i != 20 )
+    return NULL; /* Fingerprint to short.  */
+  return fpr;
+}
+
+
+
+/* CHECKCRL [<fingerprint>]
+
+   Check whether the certificate with FINGERPRINT (SHA-1 hash of the
+   entire X.509 certificate blob) is valid or not by consulting the
+   CRL responsible for this certificate.  If the fingerprint has not
+   been given or the certificate is not know, the function 
+   inquires the certficate using the
+
+      INQUIRE TARGETCERT
+
+   and the caller is expected to return the certificate for the
+   request (which should match FINGERPRINT) as a binary blob.
+   Processing then takes place without further interaction; in
+   particular dirmngr tries to locate other required certificate by
+   its own mechanism which includes a local certificate store as well
+   as a list of trusted root certificates.
+
+   The return value is the usual gpg-error code or 0 for ducesss;
+   i.e. the certificate validity has been confirmed by a valid CRL.
+*/
+static int
+cmd_checkcrl (assuan_context_t ctx, char *line)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+  gpg_error_t err;
+  unsigned char fprbuffer[20], *fpr;
+  ksba_cert_t cert;
+
+  fpr = get_fingerprint_from_line (line, fprbuffer);
+  cert = fpr? get_cert_byfpr (fpr) : NULL;
+  
+  if (!cert)
+    {
+      /* We do not have this certificate yet or the fingerprint has
+         not been given.  Inquire it from the client.  */
+      assuan_error_t ae;
+      unsigned char *value = NULL;
+      size_t valuelen; 
+      
+      ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+                           &value, &valuelen, MAX_CERT_LENGTH);
+      if (ae)
+        {
+          log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
+          err = map_assuan_err (ae);
+          goto leave;
+        }
+  
+      if (!valuelen) /* No data returned; return a comprehensible error. */
+        err = gpg_error (GPG_ERR_MISSING_CERT);
+      else
+        {
+          err = ksba_cert_new (&cert);
+          if (!err)
+            err = ksba_cert_init_from_mem (cert, value, valuelen);
+        }
+      xfree (value);
+      if(err)
+        goto leave;
+    }
+
+  assert (cert);
+
+  err = crl_cache_cert_isvalid (ctrl, cert, ctrl->force_crl_refresh);
+  if (gpg_err_code (err) == GPG_ERR_NO_CRL_KNOWN)
+    {
+      err = crl_cache_reload_crl (ctrl, cert);
+      if (!err)
+        err = crl_cache_cert_isvalid (ctrl, cert, 0);
+    }
+
+ leave:
+  if (err)
+    log_error (_("command %s failed: %s\n"), "CHECKCRL", gpg_strerror (err));
+  ksba_cert_release (cert);
+  return map_to_assuan_status (err);
+}
+
+
+/* CHECKOCSP [<fingerprint>]
+
+   Check whether the certificate with FINGERPRINT (SHA-1 hash of the
+   entire X.509 certificate blob) is valid or not by asking an OCSP
+   responder responsible for this certificate.  The optional
+   fingerprint may be used for a quick check in case an OCSP check has
+   been down for this certificate recently (we always cache OCSP
+   responses for a couple of minutes). If the fingerprint has not been
+   given or there is no cached result, the function inquires the
+   certificate using the
+
+      INQUIRE TARGETCERT
+
+   and the caller is expected to return the certificate for the
+   request (which should match FINGERPRINT) as a binary blob.
+   Processing then takes place without further interaction; in
+   particular dirmngr tries to locate other required certificates by
+   its own mechanism which includes a local certificate store as well
+   as a list of trusted root certifciates.
+
+   The return value is the usual gpg-error code or 0 for ducesss;
+   i.e. the certificate validity has been confirmed by a valid CRL.
+*/
+static int
+cmd_checkocsp (assuan_context_t ctx, char *line)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+  gpg_error_t err;
+  unsigned char fprbuffer[20], *fpr;
+  ksba_cert_t cert;
+
+  fpr = get_fingerprint_from_line (line, fprbuffer);
+  cert = fpr? get_cert_byfpr (fpr) : NULL;
+  
+  if (!cert)
+    {
+      /* We do not have this certificate yet or the fingerprint has
+         not been given.  Inquire it from the client.  */
+      assuan_error_t ae;
+      unsigned char *value = NULL;
+      size_t valuelen; 
+      
+      ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+                           &value, &valuelen, MAX_CERT_LENGTH);
+      if (ae)
+        {
+          log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
+          err = map_assuan_err (ae);
+          goto leave;
+        }
+  
+      if (!valuelen) /* No data returned; return a comprehensible error. */
+        err = gpg_error (GPG_ERR_MISSING_CERT);
+      else
+        {
+          err = ksba_cert_new (&cert);
+          if (!err)
+            err = ksba_cert_init_from_mem (cert, value, valuelen);
+        }
+      xfree (value);
+      if(err)
+        goto leave;
+    }
+
+  assert (cert);
+
+  if (!opt.allow_ocsp)
+    err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+  else
+    err = ocsp_isvalid (ctrl, cert, NULL);
+
+ leave:
+  if (err)
+    log_error (_("command %s failed: %s\n"), "CHECKOCSP", gpg_strerror (err));
+  ksba_cert_release (cert);
+  return map_to_assuan_status (err);
+}
+
+
 
 /* LOOKUP <pattern>
 
@@ -505,6 +562,7 @@ cmd_isvalid (assuan_context_t ctx, char *line)
 static int
 cmd_lookup (assuan_context_t ctx, char *line)
 {
+  ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err = 0;
   assuan_error_t ae;
   char *p;
@@ -551,7 +609,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
                    ldapserver->base?ldapserver->base : "[default]");
 
       /* Fetch certificates matching pattern */
-      err = start_cert_fetch (&fetch_context, list, ldapserver);
+      err = start_cert_fetch (ctrl, &fetch_context, list, ldapserver);
       if (err)
         {
           log_error (_("start_cert_fetch failed: %s\n"), gpg_strerror (err));
@@ -567,6 +625,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
             {
               truncated = 1;
               err = 0;
+              break;  /* Ready.  */
             }
           if (gpg_err_code (err) == GPG_ERR_EOF)
             {
@@ -637,7 +696,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
    that FILENAME should be given with an absolute path because
    Dirmngrs cwd is not known.  This command is usually used by gpgsm
    using the invocation "gpgsm --call-dirmngr loadcrl <filename>".  A
-   direct onvocation of Dirmngr is not useful because gpgsm might need
+   direct invocation of Dirmngr is not useful because gpgsm might need
    to callback gpgsm to ask for the CA's certificate.
 */
 
@@ -650,7 +709,7 @@ cmd_loadcrl (assuan_context_t ctx, char *line)
 
   buf = xmalloc (strlen (line)+1);
   strcpy_escaped_plus (buf, line);
-  err = crl_cache_load (ctrl, ctrl->server_local->crl_cache, buf);
+  err = crl_cache_load (ctrl, buf);
   xfree (buf);
   if (err)
     log_error (_("command %s failed: %s\n"), "LOADCRL", gpg_strerror (err));
@@ -670,19 +729,125 @@ static int
 cmd_listcrls (assuan_context_t ctx, char *line)
 {
   gpg_error_t err;
-  ctrl_t ctrl = assuan_get_pointer (ctx);
   FILE *fp = assuan_get_data_fp (ctx);
 
   if (!fp)
     return PARM_ERROR (_("no data stream"));
 
-  err = crl_cache_list (ctrl->server_local->crl_cache, fp);
+  err = crl_cache_list (fp);
   if (err)
     log_error (_("command %s failed: %s\n"), "LISTCRLS", gpg_strerror (err));
   return map_to_assuan_status (err);
 }
 
 
+/* CACHECERT 
+
+   Put a certificate into the internal cache.  This command might be
+   usedful if a client knows in advance certificates required for a
+   test and wnats to make sure they get added to the internal cache.
+   It is also helpful for debugging.  To get the actual certificate,
+   this command immediately inquires it using
+
+      INQUIRE TARGETCERT
+
+   and the caller is expected to return the certificate for the
+   request as a binary blob.
+*/
+static int
+cmd_cachecert (assuan_context_t ctx, char *line)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+  gpg_error_t err;
+  ksba_cert_t cert = NULL;
+  assuan_error_t ae;
+  unsigned char *value = NULL;
+  size_t valuelen; 
+      
+  ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+                       &value, &valuelen, MAX_CERT_LENGTH);
+  if (ae)
+    {
+      log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
+      err = map_assuan_err (ae);
+      goto leave;
+    }
+  
+  if (!valuelen) /* No data returned; return a comprehensible error. */
+    err = gpg_error (GPG_ERR_MISSING_CERT);
+  else
+    {
+      err = ksba_cert_new (&cert);
+      if (!err)
+        err = ksba_cert_init_from_mem (cert, value, valuelen);
+    }
+  xfree (value);
+  if(err)
+    goto leave;
+
+  err = cache_cert (cert);
+
+ leave:
+  if (err)
+    log_error (_("command %s failed: %s\n"), "CACHECERT", gpg_strerror (err));
+  ksba_cert_release (cert);
+  return map_to_assuan_status (err);
+}
+
+
+/* VALIDATE 
+
+   Validate a certificate using the certificate validationj fucntion
+   used internally by dirmngr.  This command is only useful for
+   debugging.  To get the actual certificate, this command immediately
+   inquires it using
+
+      INQUIRE TARGETCERT
+
+   and the caller is expected to return the certificate for the
+   request as a binary blob.
+*/
+static int
+cmd_validate (assuan_context_t ctx, char *line)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+  gpg_error_t err;
+  ksba_cert_t cert = NULL;
+  assuan_error_t ae;
+  unsigned char *value = NULL;
+  size_t valuelen; 
+      
+  ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+                       &value, &valuelen, MAX_CERT_LENGTH);
+  if (ae)
+    {
+      log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
+      err = map_assuan_err (ae);
+      goto leave;
+    }
+  
+  if (!valuelen) /* No data returned; return a comprehensible error. */
+    err = gpg_error (GPG_ERR_MISSING_CERT);
+  else
+    {
+      err = ksba_cert_new (&cert);
+      if (!err)
+        err = ksba_cert_init_from_mem (cert, value, valuelen);
+    }
+  xfree (value);
+  if(err)
+    goto leave;
+
+  err = validate_cert_chain (ctrl, cert, NULL, VALIDATE_MODE_CRL_RECURSIVE);
+
+ leave:
+  if (err)
+    log_error (_("command %s failed: %s\n"), "VALIDATE", gpg_strerror (err));
+  ksba_cert_release (cert);
+  return map_to_assuan_status (err);
+}
+
+
 
 /* Tell the assuan library about our commands. */
 static int
@@ -693,9 +858,13 @@ register_commands (ASSUAN_CONTEXT ctx)
     int (*handler)(ASSUAN_CONTEXT, char *line);
   } table[] = {
     { "ISVALID",    cmd_isvalid },
+    { "CHECKCRL",   cmd_checkcrl },
+    { "CHECKOCSP",  cmd_checkocsp },
     { "LOOKUP",     cmd_lookup },
     { "LOADCRL",    cmd_loadcrl },
     { "LISTCRLS",   cmd_listcrls },
+    { "CACHECERT",  cmd_cachecert },
+    { "VALIDATE",   cmd_validate },
     { "INPUT",      NULL },
     { "OUTPUT",     NULL },
     { NULL }
@@ -712,24 +881,41 @@ register_commands (ASSUAN_CONTEXT ctx)
 }
 
 
-/* Startup the server and run the main loop. */
+/* Startup the server and run the main command loop.  With FD = -1
+   used stdin/stdout. */
 void
-start_command_handler (void)
+start_command_handler (int fd)
 {
   int rc;
-  int filedes[2];
   assuan_context_t ctx;
-  struct server_control_s ctrl;
+  ctrl_t ctrl;
 
-  memset (&ctrl, 0, sizeof ctrl);
-  dirmngr_init_default_ctrl (&ctrl);
+  ctrl = xtrycalloc (1, sizeof *ctrl);
+  if (ctrl)
+    ctrl->server_local = xtrycalloc (1, sizeof *ctrl->server_local);
+  if (!ctrl || !ctrl->server_local)
+    {
+      log_error (_("can't allocate control structure: %s\n"),
+                 strerror (errno));
+      xfree (ctrl);
+      return;
+    }
+    
+  dirmngr_init_default_ctrl (ctrl);
+
+  if (fd == -1)
+    {
+      int filedes[2];
+
+      filedes[0] = 0;
+      filedes[1] = 1;
+      rc = assuan_init_pipe_server (&ctx, filedes);
+    }
+  else
+    {
+      rc = assuan_init_connected_socket_server (&ctx, fd);
+    }
 
-  /* For now we use a simple pipe based server so that we can work
-     from scripts.  We will later add options to run as a daemon and
-     wait for requests on a Unix domain socket. */
-  filedes[0] = 0;
-  filedes[1] = 1;
-  rc = assuan_init_pipe_server (&ctx, filedes);
   if (rc)
     {
       log_error (_("failed to initialize the server: %s\n"),
@@ -747,18 +933,12 @@ start_command_handler (void)
   assuan_set_hello_line (ctx, "Dirmngr " VERSION " at your service");
   assuan_register_option_handler (ctx, option_handler);
 
-  assuan_set_pointer (ctx, &ctrl);
-  ctrl.server_local = xcalloc (1, sizeof *ctrl.server_local);
-  ctrl.server_local->assuan_ctx = ctx;
-  ctrl.server_local->crl_cache = crl_cache_init ();
-  ctrl.server_local->message_fd = -1;
+  ctrl->server_local->assuan_ctx = ctx;
+  assuan_set_pointer (ctx, ctrl);
 
   if (DBG_ASSUAN)
     assuan_set_log_stream (ctx, log_get_stream ());
 
-  if (!primary_connection)
-    primary_connection = &ctrl;
-
   for (;;) 
     {
       rc = assuan_accept (ctx);
@@ -778,22 +958,29 @@ start_command_handler (void)
         }
     }
   
-  if (primary_connection == &ctrl)
-    primary_connection = NULL;
-
-  crl_cache_deinit (ctrl.server_local->crl_cache);
-
+  ldap_wrapper_connection_cleanup (ctrl);
 
+  ctrl->server_local->assuan_ctx = NULL;
   assuan_deinit_server (ctx);
+
+  if (ctrl->refcount)
+    log_error ("oops: connection control structure still referenced (%d)\n",
+               ctrl->refcount);
+  else
+    {
+      xfree (ctrl->server_local);
+      xfree (ctrl);
+    }
 }
 
 
 /* Send a status line back to the client.  KEYWORD is the status
    keyword, the optioal string argumenst are blank separated added to
    the line, the last argument must be a NULL. */
-void
+gpg_error_t
 dirmngr_status (ctrl_t ctrl, const char *keyword, ...)
 {
+  gpg_error_t err = 0;
   va_list arg_ptr;
   const char *text;
 
@@ -818,34 +1005,41 @@ dirmngr_status (ctrl_t ctrl, const char *keyword, ...)
             *p++ = *text++;
         }
       *p = 0;
-      assuan_write_status (ctx, keyword, buf);
+      err = map_assuan_err (assuan_write_status (ctx, keyword, buf));
     }
 
   va_end (arg_ptr);
+  return err;
 }
 
 
 /* Note, that we ignore CTRL for now but use the first connection to
    send the progress info back. */
-void
+gpg_error_t
 dirmngr_tick (ctrl_t ctrl)
 {
-  static clock_t next_tick = 0;
-  clock_t now = clock ();
+  static time_t next_tick = 0;
+  gpg_error_t err = 0;
+  time_t now = time (NULL);
 
   if (!next_tick)
     {
-      next_tick = now + CLOCKS_PER_SEC;
-      if (next_tick < now)
-        next_tick = 0;
+      next_tick = now + 1;
     }
   else if ( now > next_tick )
     {
-      next_tick = now + CLOCKS_PER_SEC;
-      if (next_tick < now)
-        next_tick = 0;
-      
-      if (primary_connection)
-        dirmngr_status (primary_connection, "PROGRESS", "tick", "? 0 0", NULL);
+      if (ctrl)
+        {
+          err = dirmngr_status (ctrl, "PROGRESS", "tick", "? 0 0", NULL);
+          if (err)
+            {
+              /* Take this as in indication for a cancel request.  */
+              err = gpg_error (GPG_ERR_CANCELED);
+            }
+          now = time (NULL);
+        }
+
+      next_tick = now + 1;
     }
+  return err;
 }
